/* $Id: locale.h,v 1.1 2004/09/04 06:22:17 bird Exp $ */
/** @file
 *
 * Internal InnoTek LIBC header.
 * Locale support implementation through OS/2 Unicode API.
 *
 * Copyright (c) 2003 InnoTek Systemberatung GmbH
 * Copyright (c) 2003-2004 knut st. osmundsen <bird-srcspam@anduin.net>
 *
 *
 * This file is part of InnoTek LIBC.
 *
 * InnoTek LIBC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * InnoTek LIBC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with InnoTek LIBC; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __InnoTekLIBC_locale_h__
#define __InnoTekLIBC_locale_h__

#include <sys/cdefs.h>
#include <locale.h>
#include <uconv.h>

__BEGIN_DECLS

/**
 * LC_COLLATE information.
 */
typedef struct __libc_LocaleCollate
{
    /** Character weight for SBCS codepages. */
    unsigned char   auchWeight[256];
    /** MBCS prefixes. Two bits per character. */
    unsigned char   au2MBCSPrefixs[256/4];
    /** The converter object to convert to and from selected codepage
      (used with MBCS codepages only). */
    UconvObject     uobj;
    /** The locale object. */
    LocaleObject    lobj;
    /** Non-zero if there are any MBCS prefix characters in codepage. */
    char            mbcs;
} __LIBC_LOCALECOLLATE;
/** Pointer to locale collate structure. */
typedef __LIBC_LOCALECOLLATE *__LIBC_PLOCALECOLLATE;


/**
 * This structure contains the flags and uppercase/lowercase tables.
 */
typedef struct __libc_LocaleCtype
{
    /** All uppercased characters. */
    unsigned char   auchUpper[256];
    /** All lowercased characters. */
    unsigned char   auchLower[256];
    /** Bit flags for every character (for isXXX() function series). */
    unsigned short  ausfType[256];
    /** MBCS prefixes. Two bits per character. */
    unsigned char   au2MBCSPrefixs[256/4];
    /** The converter object to convert to and from selected codepage
      (used with MBCS codepages only). */
    UconvObject     uobj;
    /** The locale object. */
    LocaleObject    lobj;
    /** Non-zero if there are any MBCS prefix characters in codepage. */
    char            mbcs;
} __LIBC_LOCALECTYPE;
/** Pointer to the Ctype locale struct. */
typedef __LIBC_LOCALECTYPE *__LIBC_PLOCALECTYPE;


/**
 * This structure contains the flags for the default (C/POSIX) locale.
 */
typedef struct __libc_LocaleCtypeDefault
{
    /** Bit flags for every character (for isXXX() function series). */
    unsigned short  ausfType[256];
} __LIBC_LOCALECTYPEDEFAULT;
/** Pointer to the Ctype locale struct. */
typedef __LIBC_LOCALECTYPEDEFAULT *__LIBC_PLOCALECTYPEDEFAULT;


/**
 * This structure keeps the time formatting rules.
 * The fConsts flag indicates what kind of memory is backing the strings.
 */
typedef struct __libc_LocaleTime
{
    /** Short month names. */
    char *smonths[12];
    /** Long month names. */
    char *lmonths[12];
    /** Short weekday names. */
    char *swdays[7];
    /** Long weekday names. */
    char *lwdays[7];
    /** Date and time format. */
    char *date_time_fmt;
    /** Date format. */
    char *date_fmt;
    /** Time format. */
    char *time_fmt;
    /** AM strings. */
    char *am;
    /** PM strings. */
    char *pm;
    /** If set all the strings are consts and shall not be free()ed. */
    int             fConsts;
} __LIBC_LOCALETIME;
/** Pointer to time locale data. */
typedef __LIBC_LOCALETIME *__LIBC_PLOCALETIME;


/**
 * Locale information structure.
 *
 * This is the lconv struct with a couple of private field indicating
 * which parts of it we have updated and assigned heap strings.
 */
typedef struct __libc_localeLconv
{
    /** The lconv structure. */
    struct lconv    s;
    /** Indicates that all the numeric members are readonly const strings. */
    int             fNumericConsts;
    /** Indicates that all the monetary members are readonly const strings. */
    int             fMonetaryConsts;
} __LIBC_LOCALELCONV;
/** Pointer to extended locale information structure. */
typedef __LIBC_LOCALELCONV *__LIBC_PLOCALELCONV;


/** String collation information. */
extern __LIBC_LOCALECOLLATE     __libc_gLocaleCollate;
/** Character case conversion tables. */
extern __LIBC_LOCALECTYPE       __libc_GLocaleCtype;
/** Character case conversion tables for the default 'C'/'POSIX' locale. */
extern const __LIBC_LOCALECTYPEDEFAULT __libc_GLocaleCtypeDefault;
/** Locale information structure. */
extern __LIBC_LOCALELCONV       __libc_gLocaleLconv;
/** Date / time formatting rules. */
extern __LIBC_LOCALETIME        __libc_gLocaleTime;


/** Convert a string to Unicode, apply some transform and convert back. */
extern void __libc_ucs2Do(UconvObject *uconv, char *s, void *arg, int (*xform)(UniChar *, void *));
/** Convert a MBCS character to Unicode; returns number of bytes in MBCS char. */
extern int  __libc_ucs2To(UconvObject, const unsigned char *, size_t, UniChar *);
/** Convert a Unicode character to MBCS. */
extern int  __libc_ucs2From(UconvObject, UniChar, unsigned char *, size_t);
/** Converts a codepage string to unichar and something libuni might recognize. */
extern void __libc_TranslateCodepage(const char *cp, UniChar *ucp);



/** Handy macros for working with the au2MBCSPrefixs members of
 * the locale data structures. The au2MBCSPrefixs members are
 * array which elements are 2 bits long.
 * @{
 */
#define SET_MBCS_PREFIX(a, c, v) \
    a[((unsigned char)(c)) >> 2] |= (v) << (2 * ((c) & 3))

#define LEN_MBCS_PREFIX(a, c) \
    ((a[((unsigned char)(c)) >> 2] >> (2 * (((c) & 3) ^ 3))) & 3)

#define IS_MBCS_PREFIX(p, c) \
    (LEN_MBCS_PREFIX((p)->au2MBCSPrefixs, c) != 1)

#define CHK_MBCS_PREFIX(p, c, v) \
    ((v = LEN_MBCS_PREFIX((p)->au2MBCSPrefixs, c)) > 1)
/** @} */

__END_DECLS

#endif /* __SYS_LOCALE_H__ */

